* Program: EdcDemo.prg
*  Action: Demonstrate the EDC class library.
*  Notice: The author releases all rights to the public domain
*        : subject to the Warranty Disclaimer in EdcLib.prg.
*  Author: Tom Rettig
* Version: EDC Version 1.0 July 15, 1995 (#defined in True.h)
*Requires: True.h and EdcLib.prg

#INCLUDE True.h

LOCAL oEDC, llCloseDatabase, llCloseClass


* Open class definition.
IF NOT "EDCLIB" $ SET("PROCEDURE")
   llCloseClass = .T.
   SET PROCEDURE TO EdcLib ADDITIVE
ENDIF

oEDC = CREATEOBJECT("EDC")

* Uncomment this to show errors as messages without a cancel option.
* =oEDC.SetShowErrors(.F.)

* Uncomment this to shut off all messages including redirected errors.
* =oEDC.SetShowMessages(.F.)

* Uncomment this to shut off parameter checking.
* =oEDC.SetTestParameters(.F.)

* Uncomment this to test performance with oEDC.nExecTime property.
=oEDC.SetExecTime(.T.)

* Open the VFP TestData database if none is open.
IF EMPTY(DBC())
   llCloseDatabase = .T.
   OPEN DATABASE HOME()+"samples\data\TestData"
ENDIF

IF oEDC.lOpen("EdcDemo")  && parameter is EDC field name (max 10 chars)
   
   ? oEDC.cVersion()

   * Test everything requiring a name with the VFP demos.
   IF ATC("TestData.dbc", DBC()) > 0

      ? oEDC.lCurrent()

      LOCAL lcObjectName, lcObjectType,;
            laAllProps[1], laWildCardProps[1]

      lcObjectName = "Orders.Cust_ID"
      lcObjectType = "Relation"
      ? oEDC.lDbcFindObject(lcObjectName, lcObjectType), oEDC.nExecTime
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_REL_TAG), oEDC.nExecTime
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_REL_FKTABLE), oEDC.nExecTime
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_REL_FKTAG), oEDC.nExecTime

      lcObjectType = "Index"
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_IND_TAGTYPE)

      lcObjectName = "Customer"
      lcObjectType = "Table"
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_TAB_FILEPATH), oEDC.nExecTime
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_TAB_PRIMARYTAG), oEDC.nExecTime

      lcObjectName = "Customer.Cust_ID"
      lcObjectType = "Index"
      ? oEDC.cDbcGetProp(lcObjectName, lcObjectType,;
                         ccVF_IND_TAGTYPE), oEDC.nExecTime

      ? oEDC.cAnalyzeDbcProp("Analyze", "table", "browseall"),;
        oEDC.nExecTime

      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyCharProp", "foo"), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyBigCharProp", REPLICATE("x",4000)), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyDateProp", DATE()), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyDateTimeProp", DATETIME()), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyRemovedProp", ""), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "NumProp", 12345.6789), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "CurProp", $12345.6789), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyZeroProp", 0000000000.000000000000), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyMinusProp", -987.65), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyNullProp", .NULL.), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyIncProp", 3), oEDC.nExecTime
      ?
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyCharProp"), oEDC.nExecTime
      ? LEFT(OEDC.uGetProp(lcObjectName, lcObjectType,;
                           "MyBigCharProp"), 10)+"...", oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyDateProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyDateTimeProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyRemovedProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "NumProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "CurProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyZeroProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyMinusProp"), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyNullProp"), oEDC.nExecTime
      ?

      FOR i = 1 TO 10
         ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                         "MyIncProp", 3), oEDC.nExecTime
         ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                         "MyIncProp"), oEDC.nExecTime
         ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                         "MyIncProp"), oEDC.nExecTime
      ENDFOR

      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncProp", 30), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncProp", -100), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp", DATE()), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp"), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp", -100), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp"), oEDC.nExecTime

      =oEDC.uSetProp(lcObjectName, lcObjectType,;
                     "My Prop With Spaces In Name", .NULL.)

      ? oEDC.uGetProp(lcObjectName, lcObjectType,;
                      "my*", @laWildCardProps),;
        oEDC.nExecTime       && must pass by reference
      DISPLAY MEMORY LIKE laWildCardProps

      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "all", @laAllProps)  && must pass by reference
      DISPLAY MEMORY LIKE laAllProps


      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncProp"), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncProp", 30), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncProp", -100), oEDC.nExecTime
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp", DATE()), oEDC.nExecTime
      ? OEDC.uIncProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp", -100), oEDC.nExecTime
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyIncDateProp"), oEDC.nExecTime
      

      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyCharProp", ccHEAD_OFF, cnHEAD_REMOVELOCK),;
        oEDC.nExecTime, oEDC.cPropHead
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyCharProp", .F., cnPROP_REMOVE),;
        oEDC.nExecTime, oEDC.cPropHead
      ? oEDC.uSetProp(lcObjectName, lcObjectType,;
                      "MyEmptyCharProp", ccHEAD_ON, cnHEAD_REMOVELOCK),;
        oEDC.nExecTime, oEDC.cPropHead

   ENDIF  && TestData open
ENDIF  && open


#if .f.
      * Test errors
      ? OEDC.uGetProp(lcObjectName, lcObjectType,;
                      "MyNonExistentProp")
      ? OEDC.uGetProp("FOO", lcObjectType,;
                      "MyNonExistentProp")
      ? OEDC.uGetProp(lcObjectName, "FOO",;
                      "MyNonExistentProp")
      ? OEDC.uGetProp(lcObjectName, "",;
                      "MyNonExistentProp")
      ? OEDC.uGetProp(lcObjectName, "FOO")

      * Test packing.
      ? oEDC.lPack("memo")
      ? oEDC.lPack("pack")
      ? oEDC.lPack("copy")

      * Test analyze.
      ? oEDC.cAnalyzeDbcProp("Analyze", "table")
      ? oEDC.cAnalyzeDbcProp("", "", "browse")
      ? oEDC.cAnalyzeDbcProp("", "", "browseall")
#endif


IF llCloseDatabase
   CLOSE DATABASE
ENDIF
IF llCloseClass
   RELEASE PROCEDURE EdcLib
ENDIF
RETURN
*** EdcDemo.prg *********************************************

