* Program: InfoLib.prg
* Classes: BaseClasses - abstract class
*        : ContainerInfo
*        : ErrorInfo
*        : MemberInfo
*        : SysInfo
*   Bases: Based on Custom
*  Notice: The author releases all rights to the public domain
*        : subject to the Warranty Disclaimer below.
*  Author: Tom Rettig
*        : Rettig Micro Corporation
*        : 2532 Lincoln Boulevard, Suite 110
*        : Marina del Rey, CA  90291-5978
*        :  Telephone: 310-301-0911
*        :        Fax: 310-821-1162
*        : CompuServe: 75066,352
*        :           : in FOXUSER forum's 3rd-Party section or MAIL
* Version: Info Version 1.0 July 15, 1995 (#defined in True.h)
*  Action: Create tables of information.
*   Usage: SET PROCEDURE TO InfoLib
*        : =CREATEOBJECT("...Info" [, cOutputFile])
*Requires: Visual FoxPro for Windows version 3.0 or later
*        : True.h named constant file (#included below)
*   Notes: - May be freely used, modified, and distributed in
*        : compiled and/or source code form.
*        : - The author appreciates acknowledgment in commercial
*        : products and publications that use or learn from this class.
*        : - Technical support is not officially provided.  The
*        : author is very interested in hearing about problems
*        : or enhancement requests you have, and will try to be
*        : helpful within reasonable limits.  Email or fax preferred.
*        : - Warranty Disclaimer: NO WARRANTY!!!
*        : THE AUTHOR RELEASES TO THE PUBLIC DOMAIN ALL CLAIMS TO ANY
*        : RIGHTS IN THIS PROGRAM AND FREELY PROVIDES IT AS IS WITHOUT
*        : WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING, BUT NOT
*        : LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*        : FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR, OR ANY
*        : OTHER PARTY WHO MAY MODIFY AND/OR REDISTRIBUTE THIS PROGRAM, BE
*        : LIABLE FOR ANY COMMERCIAL, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
*        : DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM
*        : INCLUDING, BUT NOT LIMITED TO, LOSS OF DATA OR DATA BEING
*        : RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR LOSSES
*        : SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE PROGRAM TO
*        : OPERATE WITH ANY OTHER PROGRAMS, EVEN IF YOU OR OTHER PARTIES
*        : HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

#INCLUDE True.h

DEFINE CLASS BaseClasses AS Custom  && abstract class
   PROTECTED aBaseClass[32]
   aBaseClass[ 1] = "CheckBox"
   aBaseClass[ 2] = "Column"
   aBaseClass[ 3] = "ComboBox"
   aBaseClass[ 4] = "CommandButton"
   aBaseClass[ 5] = "CommandGroup"
   aBaseClass[ 6] = "Container"
   aBaseClass[ 7] = "Control"
   aBaseClass[ 8] = "Cursor"
   aBaseClass[ 9] = "Custom"
   aBaseClass[10] = "DataEnvironment"
   aBaseClass[11] = "EditBox"
   aBaseClass[12] = "Form"
   aBaseClass[13] = "FormSet"
   aBaseClass[14] = "Grid"
   aBaseClass[15] = "Header"
   aBaseClass[16] = "Image"
   aBaseClass[17] = "Label"
   aBaseClass[18] = "Line"
   aBaseClass[19] = "ListBox"
   aBaseClass[20] = "OleControl"
   aBaseClass[21] = "OleBoundControl"
   aBaseClass[22] = "OptionButton"
   aBaseClass[23] = "OptionGroup"
   aBaseClass[24] = "Page"
   aBaseClass[25] = "PageFrame"
   aBaseClass[26] = "Relation"
   aBaseClass[27] = "Separator"
   aBaseClass[28] = "Shape"
   aBaseClass[29] = "Spinner"
   aBaseClass[30] = "TextBox"
   aBaseClass[31] = "Timer"
   aBaseClass[32] = "ToolBar"
ENDDEFINE  && CLASS BaseClasses AS Custom  && abstract class


DEFINE CLASS ContainerInfo AS BaseClasses
   * Test which objects can contain which other objects.
   * Press Escape when presented with the OleControl dialog.

   PROTECTED cOutputDefault,;
             nOutputSize
   cOutputDefault = "InfoCon"
   nOutputSize    = 50000

   PROTECTED PROCEDURE Init(tcOutput)
      LOCAL lnCounter,;
            lnCounter2,;
            loObject,;
            lcOutput,;
            llStatusBar,;
            llTalk

      IF DISKSPACE() < THIS.nOutputSize
         ERROR cnVF_ERR_DISK_SPACE, "output file."
         RETURN .F.  && early exit
      ENDIF

      * Create table to store results.
      lcOutput = IIF(EMPTY(tcOutput), THIS.cOutputDefault, tcOutput)
      CREATE TABLE (lcOutput) FREE;
         (cClass     C(15),;
          cContains  C(15),;
          cError     C(60))
      IF NOT (USED() OR ALIAS() $ UPPER(lcOutput))
         RETURN .F.  && early exit
      ENDIF

      * Environment.
      llTalk = SET("TALK")=="ON"
      SET TALK OFF
      llStatusBar = SET("STATUS BAR")=="ON"
      SET STATUS BAR ON

      * Test each base class.
      FOR lnCounter = 1 TO ALEN(THIS.aBaseClass)
         IF INLIST(lnCounter, 20, 21)  && skip ole controls
            LOOP
         ENDIF

         loObject = CREATEOBJECT(THIS.aBaseClass[lnCounter])
         FOR lnCounter2 = 1 TO ALEN(THIS.aBaseClass)
            SET MESSAGE TO THIS.aBaseClass[lnCounter] + "."+;
                           THIS.aBaseClass[lnCounter2]
            INSERT INTO (ALIAS());
                        (cClass, cContains);
                 VALUES (THIS.aBaseClass[lnCounter],;
                         THIS.aBaseClass[lnCounter2])

            ON ERROR REPLACE cContains WITH IIF(ERROR() == 1734,;
                                                 "", cContains),;
                             cError WITH MESSAGE()+" "+LTRIM(STR(ERROR()))
            loObject.AddObject(THIS.aBaseClass[lnCounter2],;
                               THIS.aBaseClass[lnCounter2])
            ON ERROR
            IF RIGHT(TRIM(cERROR), 4) == "1734"
               EXIT  && no AddObject method
            ENDIF
         ENDFOR
      ENDFOR
      SET MESSAGE TO

      * Index all fields.
      FOR lnCounter = 1 TO FCOUNT()
         lcOutput = FIELD(lnCounter)
         INDEX ON &lcOutput TAG (FIELD(lnCounter))  && macro alert
      ENDFOR
      SET ORDER TO

      * Clean up but leave table open for browse.
      IF llTalk
         SET TALK ON
      ENDIF
      IF NOT llStatusBar
         SET STATUS BAR OFF
      ENDIF

      BROWSE LAST
      RETURN .F.
   ENDPROC  && Init
ENDDEFINE  && CLASS ContainerInfo AS BaseClasses


DEFINE CLASS ErrorInfo AS Custom
   PROTECTED nErrorInvalid,;
             nErrorBegin,;
             nErrorEnd,;
             cOutputDefault,;
             nOutputSize
   nErrorInvalid  = 0
   nErrorBegin    = 1
   nErrorEnd      = 2000
   cOutputDefault = "InfoErr"
   nOutputSize    = 100000

   PROTECTED PROCEDURE Init(tcOutput)
      LOCAL lnCounter,;
            lcMessage,;
            lcMessageInvalid,;
            lcOutput,;
            llStatusBar,;
            llTalk

      IF DISKSPACE() < THIS.nOutputSize
         ERROR cnVF_ERR_DISK_SPACE, "output file."
         RETURN .F.  && early exit
      ENDIF

      * Create table to store results.
      lcOutput = IIF(EMPTY(tcOutput), THIS.cOutputDefault, tcOutput)
      CREATE TABLE (lcOutput) FREE;
         (nError   N(  4),;
          cMessage C(130))
      IF NOT (USED() OR ALIAS() $ UPPER(lcOutput))
         RETURN .F.  && early exit
      ENDIF

      * Environment.
      llTalk = SET("TALK")=="ON"
      SET TALK OFF
      llStatusBar = SET("STATUS BAR")=="ON"
      SET STATUS BAR ON

      ON ERROR *   && prevent generated errors from interfering

      * Get the invalid error message.
      SET MESSAGE TO "Error 0/" + LTRIM(STR(THIS.nErrorEnd))
      ERROR THIS.nErrorInvalid
      lcMessageInvalid = MESSAGE()
      INSERT INTO (ALIAS());
                  (nError, cMessage);
           VALUES (THIS.nErrorInvalid, lcMessageInvalid)

      * Test all errors.
      FOR lnCounter = THIS.nErrorBegin TO THIS.nErrorEnd
         IF lnCounter%10 == 0
            SET MESSAGE TO "Error " + LTRIM(STR(lnCounter))+;
                                "/" + LTRIM(STR(THIS.nErrorEnd))
         ENDIF
         ERROR lnCounter
         lcMessage = MESSAGE()
         IF NOT lcMessage == lcMessageInvalid
            INSERT INTO (ALIAS());
                        (nError, cMessage);
                 VALUES (lnCounter, lcMessage)
         ENDIF
      ENDFOR
      GO TOP
      SET MESSAGE TO

      * Clean up but leave table open for browse.
      ON ERROR
      IF llTalk
         SET TALK ON
      ENDIF
      IF NOT llStatusBar
         SET STATUS BAR OFF
      ENDIF

      BROWSE LAST
      RETURN .F.
   ENDPROC  && Init
ENDDEFINE  && CLASS ErrorInfo AS Custom


DEFINE CLASS MemberInfo AS BaseClasses
   * Get all built-in properties and methods into a table.
   * Test properties for read-only.

   PROTECTED cOutputDefault,;
             nOutputSize
   cOutputDefault = "InfoMem"
   nOutputSize    = 175000

   PROTECTED PROCEDURE Init(tcOutput)
      LOCAL lnCounter,;
            lnCounter2,;
            loObject,;
            lcOutput,;
            llStatusBar,;
            llTalk

      IF DISKSPACE() < THIS.nOutputSize
         ERROR cnVF_ERR_DISK_SPACE, "output file."
         RETURN .F.  && early exit
      ENDIF

      * Create table to store results.
      lcOutput = IIF(EMPTY(tcOutput), THIS.cOutputDefault, tcOutput)
      CREATE TABLE (lcOutput) FREE;
         (cBaseClass  C(15),;
          cMember     C(10),;
          cName       C(40),;
          lReadOnly   L)
      IF NOT (USED() OR ALIAS() $ UPPER(lcOutput))
         RETURN .F.  && early exit
      ENDIF

      * Environment.
      llTalk = SET("TALK")=="ON"
      SET TALK OFF
      llStatusBar = SET("STATUS BAR")=="ON"
      SET STATUS BAR ON

      * Test each base class.
      FOR lnCounter = 0 TO ALEN(THIS.aBaseClass)
         IF INLIST(lnCounter, 20, 21)  && skip ole controls
            LOOP
         ENDIF

         * Zero is _SCREEN.
         IF lnCounter == 0
            loObject = _SCREEN
            SET MESSAGE TO "_SCREEN"
         ELSE
            loObject = CREATEOBJECT(THIS.aBaseClass[lnCounter])
            SET MESSAGE TO loObject.BaseClass
         ENDIF

         DIMENSION laMembers[1, 2]
         laMembers = ""
         =AMEMBERS(laMembers, loObject, 1)
         FOR lnCounter2 = 1 TO ALEN(laMembers, 1)
            IF lnCounter == 0
               SET MESSAGE TO "_SCREEN." + laMembers[lnCounter2, 1]
            ELSE
               SET MESSAGE TO THIS.aBaseClass[lnCounter] + "."+;
                              laMembers[lnCounter2, 1]
            ENDIF
            INSERT INTO (ALIAS());
                        (cBaseClass, cMember, cName);
                 VALUES (IIF(lnCounter == 0,;
                             "_SCREEN",;
                             loObject.BaseClass),;
                         laMembers[lnCounter2, 2],;
                         laMembers[lnCounter2, 1])
            * Test properties for read-only.
            IF laMembers[lnCounter2, 2] == "Property" AND;
                  NOT TYPE("loObject." + laMembers[lnCounter2,1]) $ "UO"
               ON ERROR REPLACE lReadOnly WITH .T.
               STORE EVALUATE("loObject." + laMembers[lnCounter2,1]);
                  TO ("loObject.laMembers[lnCounter2,1]")
               ON ERROR
            ENDIF
         ENDFOR
      ENDFOR
      SET MESSAGE TO

      * Index all fields.
      FOR lnCounter = 1 TO FCOUNT()
         lcOutput = FIELD(lnCounter)
         INDEX ON &lcOutput TAG (FIELD(lnCounter))  && macro alert
      ENDFOR
      SET ORDER TO

      * Clean up but leave table open for browse.
      IF llTalk
         SET TALK ON
      ENDIF
      IF NOT llStatusBar
         SET STATUS BAR OFF
      ENDIF

      BROWSE LAST
      RETURN .F.
   ENDPROC  && Init
ENDDEFINE  && CLASS MemberInfo AS BaseClasses


DEFINE CLASS SysInfo AS Custom
* Just a wrapper class to call the SysTest program.
   PROTECTED PROCEDURE Init(tcMethod, tcOutput)
      LOCAL lcFile
      lcFile = LOCFILE("SysTest.prg", "prg",;
                       IIF("Windows 4." $ OS(),;
                           "Systest.prg",;
                           "Where is Systest.prg?"))
      IF NOT EMPTY(lcFile)
         DO (lcFile) WITH tcMethod, tcOutput
      ENDIF
      RETURN .F.
   ENDPROC  && Init
ENDDEFINE  && CLASS SysInfo AS Custom

*** Infolib.prg *********************************************
